# Copyright French Alternative Energies and Atomic Energy Commission
# Contributors: valjean developers
# valjean-support@cea.fr
#
# This software is a computer program whose purpose is to analyze and
# post-process numerical simulation results.
#
# This software is governed by the CeCILL license under French law and abiding
# by the rules of distribution of free software. You can use, modify and/ or
# redistribute the software under the terms of the CeCILL license as circulated
# by CEA, CNRS and INRIA at the following URL: http://www.cecill.info.
#
# As a counterpart to the access to the source code and rights to copy, modify
# and redistribute granted by the license, users are provided only with a
# limited warranty and the software's author, the holder of the economic
# rights, and the successive licensors have only limited liability.
#
# In this respect, the user's attention is drawn to the risks associated with
# loading, using, modifying and/or developing or reproducing the software by
# the user in light of its specific status of free software, that may mean that
# it is complicated to manipulate, and that also therefore means that it is
# reserved for developers and experienced professionals having in-depth
# computer knowledge. Users are therefore encouraged to load and test the
# software's suitability as regards their requirements in conditions enabling
# the security of their systems and/or data to be ensured and, more generally,
# to use and operate it in the same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

'''Module containing all available methods to convert a test result in a table
to be converted in rst.
'''
import logging
from collections import OrderedDict
import numpy as np
from ..cosette.task import TaskStatus
from ..gavroche.diagnostics.stats import TestOutcome, classification_counts
from .templates import PlotTemplate, CurveElements, SubPlotElements, join
from .verbosity import Verbosity


# turn off pylint warnings about invalid names in this file; there are just too
# many long function names and they cannot be renamed because
# javert.representation looks for them by programmatically constructing their
# name based on the name of the test result class, the verbosity, etc.
# pylint: disable=invalid-name


LOGGER = logging.getLogger(__name__)


def dimensions_from_array(array_shape):
    '''Check if array is consistent with 1D plot.

    Checks are done on presence of non-trivial and trivial dimensions, trivial
    being a dimension with size equal to one, non-trivial being for a dimension
    with at least two values.

    :param tuple(int) array_shape: shape of the **values** array
    :returns: indices of the non-trivial dimensions in the shape tuple
    :rtype: tuple(int)
    '''
    non_trivial_dims = tuple(d for d, s in enumerate(array_shape) if s > 1)
    LOGGER.debug("Non-trivial dimensions: %s", non_trivial_dims)
    if not non_trivial_dims:
        LOGGER.debug("Only trivial dimensions, you may prefer a different "
                     "kind of plot (PlotPoint).")
        return None
    return non_trivial_dims


def dimensions_and_bins(bins, array_shape):
    '''Determine the dimensions of the result from the
    :obj:`collections.OrderedDict` of bins. It is expected to be the "only
    non-trivial" dimensions.

    :param bins: bins coming from the results dataset
    :type bins: collections.OrderedDict(str, numpy.ndarray)
    :param tuple(int) array_shape: shape of the **values** array from the
        results dataset
    :returns: dimensions and bins to be used (non-trivial ones), or None if all
      the dimensions are trivial
    :rtype: collections.OrderedDict
    '''
    idims = dimensions_from_array(array_shape)
    if idims is None:
        return None
    if not bins:
        return None
    subd = OrderedDict(pair for i, pair in enumerate(bins.items())
                       if i in idims)
    return subd


def trim_range(bins, max_ratio=1000):
    '''Adapt bins range when extreme bins are very large.

    This function suggests reasonable ranges for the given bin axes by trimming
    the extreme bins if they are too wide. For each axis, the extreme bins are
    trimmed if their width is at least `max_ratio` times larger than the width
    of the neighbouring bin (the first bin is compared to the second one, and
    the last bin is compared to the second-last one). If there is no need to
    change the previous limits a tuple with initial limits is returned for the
    considered dimension. A boolean is associated to the tuple to precise if
    the limits have be changed.

    The trimmed ranges extend a bit over the clipped bins, so that their
    content is still visible in the plots.

    :param list(numpy.ndarray) bins: bins of the :class:`~.PlotTemplate`.
    :param float max_ratio: the bin size ratio above which bins will be
        truncated.
    :returns: new limits for all dimensions. The list has one item per
        dimension; the bool indicates whether the axis was actually trimmed.
    :rtype: list(tuple(float, float, bool))

    Examples:

    >>> bins = [np.array([-1e4, 0, 2, 4, 1e4]),
    ...         np.array([-1e2, 0, 2, 4, 1e4]),
    ...         np.array([-1e2, 0, 2, 4, 1e2])]
    >>> trim_range(bins)
    [(-0.2, 4.2, True), (-100.0, 9.2, True), (-100.0, 100.0, False)]
    '''
    LOGGER.debug('In plot_repr.trim_range')
    blimits = []
    for lbins in bins:
        nbins = lbins
        has_changed = [False, False]
        limits = [nbins[0], nbins[-1]]
        if lbins.dtype.kind == 'U':
            blimits.append(tuple(limits + [False]))
            continue
        if nbins.size < 3:
            LOGGER.note('Not enough bins to adapt range.')
            blimits.append(tuple(limits + [False]))
            continue
        if nbins.size < 4:
            LOGGER.note('Will adapt range for %d bins, '
                        'binning might be not suitable', nbins.size)
        binw = np.ediff1d(lbins)
        if binw[0]/binw[1] > max_ratio:
            limits[0] = nbins[1]
            has_changed[0] = True
        if binw[-1]/binw[-2] > max_ratio:
            limits[1] = nbins[-2]
            has_changed[1] = True
        rwidth = limits[1] - limits[0]
        if has_changed[0] or len(bins) == 1:
            limits[0] -= 0.05 * rwidth
        if has_changed[1] or len(bins) == 1:
            limits[1] += 0.05 * rwidth
        blimits.append((limits[0], limits[1], any(has_changed)))
    return blimits


def fit_curve_ranges(curves, threshold=0.0):
    '''Return a set of best-fit limits for the given curves.

    This function suggests axis limits for the given curves that make all
    curves fill the plot area. The limits are determined by looking at the
    region where any of the curves exceed the given threshold.

    :param list(CurveElements) curves: list of curves
    :param float threshold: threshold for the detection
    :returns: suggested limits for all dimensions
    :rtype: list((float, float))

    Example:

    >>> bins = [np.array([-1.0, 0.0, 1.0]),
    ...         np.array([0.0, 5.0, 10.0, 15.0])]
    >>> curve1 = CurveElements(np.array([[1.0, 2.0, 3.0],
    ...                                  [2.0, 3.0, 4.0]]),
    ...                       bins=bins, legend='curve1')
    >>> curve2 = CurveElements(np.array([[1.0, 4.0, 3.0],
    ...                                  [2.0, 3.0, 2.0]]),
    ...                       bins=bins, legend='curve2')
    >>> fit_curve_ranges([curve1, curve2], threshold = 3.5)
    [(-1.0, 1.0), (5.0, 15.0)]
    '''
    if not curves:
        return []
    c_limits = np.array([curve_limits(curve, threshold) for curve in curves])
    # the shape of c_limits here is (n_curves, n_axes, 2)
    unions = ranges_union(c_limits)
    # the shape of unions here is (n_axes, 2)
    return [(union[0], union[1]) for union in unions]


def ranges_union(ranges, union_axis=0, minmax_axis=-1):
    '''Return the union of the given ranges, axis by axis.

    :param ranges: an n-dimensional array of range bounds. The array shape can
        be anything, as long as one of the dimensions has length 2; this axis
        is assumed to contain the ``(min_range, max_range)`` pair.
    :type ranges: list or numpy.ndarray
    :param union_axis: the axis along which the union will be computed. If
        unspecified, the first axis will be assumed.
    :type union_axis: anything that `NumPy` understands as an axis.
    :param minmax_axis: the length-2 axis along which the range minimum and
        maximum are stored. If unspecified, the last axis will be assumed.
    :type minmax_axis: int
    :returns: an array of union bounds, with the same shape as the input array
        except for the suppression of `union_axis`.
    :rtype: numpy.ndarray

    Examples:

    >>> ranges_union([[(-1, 1)]])
    array([[-1,  1]])

    >>> ranges_union([[(-5, 5)], [(-3, 8)]])
    array([[-5,  8]])

    >>> ranges_union([[(-5, 5), (-1, 1)],
    ...               [(-3, 8), (-2, 2)]])
    array([[-5,  8],
           [-2,  2]])

    >>> ranges_union([[(-5, 5), (-1, 1)],
    ...               [(-3, 8), (-2, 2)],
    ...               [(-7, 1), (0, 6)]])
    array([[-7,  8],
           [-2,  6]])
    '''
    if not isinstance(ranges, np.ndarray):
        ranges = np.array(ranges)
    min_limits = np.min(ranges.take(0, minmax_axis), axis=union_axis)
    max_limits = np.max(ranges.take(1, minmax_axis), axis=union_axis)
    result = np.stack((min_limits, max_limits), axis=minmax_axis)
    return result


def curve_limits(curve, threshold=0.0):
    '''Return the set of bounds over each axis where the given curve exceeds
    the given threshold.

    For example, the following curve exceeds the threshold value of 50 in the
    central bins, which corresponds to `2.0<x<4.0` and `-0.5<y<0.5`:

    >>> bins = [np.linspace(0.0, 6.0, num=7),
    ...         np.linspace(-2.5, 2.5, num=6)]
    >>> curve = CurveElements(np.array([[  1,   1,   1,   1,   1],
    ...                                 [  1,  10,  30,  10,   1],
    ...                                 [  1,  10, 100,  10,   1],
    ...                                 [  1,  10, 100,  10,   1],
    ...                                 [  1,  10,  30,  10,   1],
    ...                                 [  1,   1,   1,   1,   1]]),
    ...                      bins=bins, legend='curve')
    >>> curve_limits(curve, threshold=50)
    [(2.0, 4.0), (-0.5, 0.5)]

    Changing the threshold to 20 modifies the range for the first axis, but
    not for the second one.

    >>> curve_limits(curve, threshold=20)
    [(1.0, 5.0), (-0.5, 0.5)]

    The function also works if the number of bin edges is equal to the number
    of values along a given direction (as opposed to the number of values plus
    one):

    >>> bins = [np.array([-1.0, 1.0]), np.array([0.0, 5.0, 10.0, 15.0])]
    >>> curve = CurveElements(np.array([[1.0, 2.0, 3.0],
    ...                                 [2.0, 3.0, 4.0]]),
    ...                      bins=bins, legend='curve')
    >>> curve_limits(curve, threshold = 3.5)
    [(1.0, 1.0), (10.0, 15.0)]
    '''
    limits = []
    n_dims = len(curve.bins)
    mask_nd = curve.values > threshold
    LOGGER.debug('mask_nd: %s', mask_nd)
    for i_bin, a_bin in enumerate(curve.bins):
        axis = tuple(i for i in range(n_dims) if i != i_bin)
        mask = mask_nd.any(axis=axis)
        LOGGER.debug('mask: %s', mask)
        if len(a_bin) == len(mask):
            over_thr_min = a_bin[mask]
            over_thr_max = a_bin[mask]
        else:  # here len(mask) + 1 == len(a_bin)
            over_thr_min = a_bin[:-1][mask]
            over_thr_max = a_bin[1:][mask]
        limits.append((np.amin(over_thr_min), np.amax(over_thr_max)))
    return limits


def pad_range(limits, log, padding=0.05):
    '''Pad the given limits.

    This function adds a bit of padding to the input limit range. If `log` is
    `False`, the new limits will be ``(limits[0] - padding * delta, limits[1] +
    padding * delta)``, where ``delta = limits[1] - limits[0]``. In log scale
    (`log=True`), the padded limits are ``(limits[0] * exp(-padding *
    log_delta), limits[1] * exp(padding * log_delta)``, with ``log_delta =
    log(limits[1] / limits[0])``.

    :param (float, float) limits: a pair of floats.
    :param bool log: whether we are in log scale.
    :param float padding: the amount of padding to insert.
    :returns: new limits
    :rtype: (float, float)
    '''
    if log:
        delta = np.exp(0.5 * padding * np.log(limits[1] / limits[0]))
        return limits[0] / delta, limits[1] * delta
    delta = 0.5 * padding * (limits[1] - limits[0])
    return limits[0] - delta, limits[1] + delta


def post_treatment(templates, result):
    '''Post-treatment of plots after template generate.

    For example add names from test result if not done before, suppress zero
    bins at range edges, etc.
    '''
    LOGGER.debug('calling post_treatment for %s', result.__class__)
    if 'Bonferroni' in result.__class__.__name__:
        LOGGER.debug('post already applied')
        return templates
    for templ in templates:
        for splt in templ.subplots:
            if splt.ptype not in ('1D', '2D'):
                continue
            blimits = [trim_range(sc.bins) for sc in splt.curves]
            if all(not b[2] for a in blimits for b in a):
                continue
            nlimits = []
            for idim in range(len(blimits[0])):
                nlimits.append((min(crv[idim][0] for crv in blimits),
                                max(crv[idim][1] for crv in blimits)))
            splt.attributes.limits = nlimits
    return templates


def build_plot_template_with_dim(curves, axnames):
    '''Organise curves in :class:`~.templates.SubPlotElements`.

    If curves are 1D they will all go in the same
    :class:`~.templates.SubPlotElements`, if there are in 2 dimensions each
    curve will have its :class:`~.templates.SubPlotElements`.

    :param list(CurveElements) curves: list of curves
    :param list(str) axnames: list of axis names
    :rtype: list(SubPlotElements)
    '''
    dim = curves[0].values.ndim
    if dim == 1:
        subplot = SubPlotElements(curves=curves, axnames=axnames, ptype='1D')
        return PlotTemplate(subplots=[subplot], small_subplots=True,
                            suppress_xaxes=True, suppress_legends=True)
    if dim == 2:
        subplots = [SubPlotElements(curves=[crv], axnames=axnames, ptype='2D')
                    for crv in curves]
        return PlotTemplate(subplots=subplots, small_subplots=False)
    LOGGER.warning('Not expected dimension')
    return []


def repr_datasets_values(result):
    '''Representation of the datasets values from test results obtained from a
    child test of :class:`~valjean.gavroche.test.TestDataset`.

    Examples:

    * :class:`~valjean.gavroche.test.TestResultEqual`;
    * :class:`~valjean.gavroche.test.TestResultApproxEqual`;
    * :class:`~valjean.gavroche.stat_tests.student.TestResultStudent`.

    :param TestResult result: test result from test on datasets
    :rtype: list(PlotTemplate)

    If the dimension cannot be determined or if the dimension is greater than 2
    an empty list is returned.
    '''
    dab = dimensions_and_bins(result.test.dsref.bins,
                              result.test.dsref.value.shape)
    if dab is None:
        return []
    if len(dab) > 2:
        LOGGER.info('No plot available for %dD, no PlotTemplate built',
                    len(dab))
        return []
    cds = [CurveElements(
        values=result.test.dsref.value, bins=list(dab.values()),
        legend=(result.test.dsref.name if result.test.dsref.name
                else 'reference'),
        index=0, errors=result.test.dsref.error)]
    cds.extend(
        CurveElements(values=ds.value, bins=list(dab.values()),
                      legend=(ds.name if ds.name else 'dataset '+str(ids)),
                      index=ids+1, errors=ds.error)
        for ids, ds in enumerate(result.test.datasets))
    pltemp = build_plot_template_with_dim(
        cds, axnames=list(dab.keys())+[result.test.dsref.what])
    return [pltemp] if pltemp else []


def repr_testresultequal(result, _verbosity=Verbosity.DEFAULT):
    '''Represent the equal test result as a plot.

    :param TestResultEqual result: a test result.
    :rtype: list(PlotTemplate)
    '''
    return repr_datasets_values(result)


def repr_testresultapproxequal(result, _verbosity=Verbosity.DEFAULT):
    '''Represent the approx equal test result as a plot.

    :param TestResultApproxEqual result: a test result.
    :rtype: list(PlotTemplate)
    '''
    return repr_datasets_values(result)


def repr_testresultstudent(result, verbosity=Verbosity.DEFAULT):
    '''Plot the Student results according to verbosity.

    :param TestResultStudent result: a test result.
    :returns: Representation of a :class:`~.TestResultStudent` as a plot.
    :rtype: list(PlotTemplate)
    '''
    LOGGER.debug("PLOT student repr, %s, res = %s", verbosity, bool(result))
    # faire une condition sur le bool dans le cas de SUMMARY ?
    if verbosity in (Verbosity.SILENT, Verbosity.SUMMARY):
        return []
    if verbosity in (Verbosity.DEFAULT, Verbosity.INTERMEDIATE):
        return repr_student_intermediate(result)
    if verbosity.value >= Verbosity.FULL_DETAILS.value:
        return repr_student_full_details(result)
    return repr_student_intermediate(result)


def repr_student_intermediate(result):
    '''Represent the Student test result as a plot.

    By default two :class:`~.templates.PlotTemplate` are returned in order to
    get a top plot representing the two series of values and the bottom plot
    representing Student's t-statistic.

    :param TestResultStudent result:  a test result.
    :rtype: list(PlotTemplate)
    '''
    rval = repr_datasets_values(result)
    rtstud = repr_student_tstud(result)
    rstudent = [join(rvals, rtstud)
                for rvals, rtstud in zip(rval, rtstud)]
    if not rstudent and rval:
        rstudent = rval
    return rstudent


def repr_student_full_details(result):
    '''Represent the Student test result as a plot.

    By default two :class:`~.templates.PlotTemplate` are returned in order to
    get a top plot representing the two series of values and the bottom plot
    representing Student's t-statistic.

    :param TestResultStudent result:  a test result.
    :rtype: list(PlotTemplate)
    '''
    LOGGER.debug('in repr_student_full_details')
    rval = repr_datasets_values(result)
    rtstud = repr_student_tstud(result)
    rpval = repr_student_pvalues(result)
    rstudent = [join(rvals, rtstud, rpvals)
                for rvals, rtstud, rpvals in zip(rval, rtstud, rpval)]
    if not rstudent and rval:
        rstudent = [join(rvals, rtstud) for rvals, rtstud in zip(rval, rtstud)]
    return rstudent


def repr_student_tstud(result):
    '''Represent the t distribution from a Student test result as a plot.

    :param TestResultStudent result: a test result.
    :rtype: list(PlotTemplate)

    .. note::
        if we have a member ``units`` in dataset the axis names would be
        constructed like ``name + units['name']``.

    If the dimension cannot be determined an empty list is returned.
    '''
    dab = dimensions_and_bins(result.test.dsref.bins,
                              result.test.dsref.value.shape)
    if dab is None:
        return []
    if len(dab) > 2:
        LOGGER.info('No plot available for %dD, no PlotTemplate built',
                    len(dab))
        return []
    curves = [CurveElements(values=tstud, legend='', bins=list(dab.values()),
                            index=ind+1, errors=None)
              for ind, tstud in enumerate(result.tstud)]
    pltemp = build_plot_template_with_dim(
        curves, axnames=list(dab.keys()) + [r'$t_{Student}$'])
    if not pltemp:
        return []
    if result.tstud[0].ndim == 1:
        for splt in pltemp.subplots:
            splt.attributes.lines = [{'y': result.test.threshold},
                                     {'y': -result.test.threshold}]
    return [pltemp]


def repr_student_pvalues(result):
    '''Representation of p-values from Student test result as a plot.

    :param TestResultStudent result: a test result.
    :rtype: list(PlotTemplate)

    If p-values were not calculated, no :class:`~.PlotTemplate` is built, so
    an empty list is returned.
    '''
    if result.pvalue is None:
        LOGGER.debug("p-value is None, won't be possible to plot it.")
        return []
    dab = dimensions_and_bins(result.test.dsref.bins,
                              result.test.dsref.value.shape)
    if dab is None:
        return []
    if len(dab) > 2:
        LOGGER.info('No plot available for %dD, no PlotTemplate built',
                    len(dab))
        return []
    curves = [CurveElements(values=pval, bins=list(dab.values()),
                            legend='', index=ind+1, errors=None)
              for ind, pval in enumerate(result.pvalue)]
    pltemp = build_plot_template_with_dim(
        curves, axnames=list(dab.keys())+['p-value'])
    if not pltemp:
        return []
    if result.pvalue[0].ndim == 1:
        for splt in pltemp.subplots:
            splt.attributes.lines = [{'y': result.test.alpha/2},
                                     {'y': -result.test.alpha/2}]
    return [pltemp]


def repr_testresultmetadata(_result, _verbosity=Verbosity.DEFAULT):
    '''Plot metadata test -> no plot done.

    :returns: empty list
    '''
    return []


def repr_testresultexternal(_result, _verbosity=Verbosity.DEFAULT):
    '''Plot external test -> no plot done.

    If plots are required they are already done. Their representation in the
    report is done by :class:`.ExternalRepresenter`.

    :returns: empty list
    '''
    return []


def repr_testresultfailed(_result, _verbosity=Verbosity.DEFAULT):
    '''Plot failed test -> no plot done.

    :returns: empty list
    '''
    return []


def repr_testresultstats(result, status_ok, label):
    '''Represent result from statistical test as a plot (pie chart probably).

    Null results are omitted from the table.

    :param TestResult result: result from a statistical test
    :rtype: list(PlotTemplate)
    '''
    statuses, counts = classification_counts(result.classify, status_ok)
    statuses_txt = [status.name for status in statuses]
    subplot = SubPlotElements(
        curves=[CurveElements(values=np.array(counts),
                              bins=[np.array(statuses_txt)],
                              legend='')],
        axnames=(label, ''), ptype='pie')
    return [PlotTemplate(subplots=[subplot], small_subplots=False)]


def repr_testresultstatstasks(result, _verbosity=Verbosity.DEFAULT):
    '''Represent result from statistical test on tasks as a plot (pie chart).

    :param TestResultStatsTasks result: result from tasks statistics
    :rtype: list(PlotTemplate)
    '''
    return repr_testresultstats(result, TaskStatus.DONE, 'Tasks')


def repr_testresultstatstests(result, _verbosity=Verbosity.DEFAULT):
    '''Represent result from statistical test on tests as a plot (pie chart).

    :param TestResultStatsTests result: result from tasks statistics
    :rtype: list(PlotTemplate)
    '''
    return repr_testresultstats(result, TestOutcome.SUCCESS, 'Tests')


def repr_statstestsby2labels(result):
    '''Plot statistics on the tests classified by 2 labels.

    Caution: the tests are summed over the second label, only the first one
    will be written on the plot. This is for summaries.

    :rtype: list(PlotTemplate)
    '''
    keys = sorted(set(res['labels'][0] for res in result.classify))
    dok, dko = (OrderedDict([(k, 0) for k in keys]) for i in range(2))
    for res in result.classify:
        if res['KO'] == 0:
            dok[res['labels'][0]] += 1
        else:
            dko[res['labels'][0]] += 1
    labs = result.test.by_labels
    yname = f'# of {labs[1]} in {labs[0]}'
    curve_ok = CurveElements(np.array(list(dok.values())), bins=[keys],
                             index=0, legend='OK')
    curve_ko = CurveElements(np.array(list(dko.values())), bins=[keys],
                             index=1, legend='KO')
    sbpeb = SubPlotElements(curves=[curve_ok, curve_ko],
                            axnames=[labs[0], yname], ptype='bar')
    sbpebs = SubPlotElements(curves=[curve_ok, curve_ko],
                             axnames=[labs[0], yname], ptype='barstack')
    return [PlotTemplate(subplots=[sbpeb, sbpebs], small_subplots=False)]


def repr_testresultstatstestsbylabels(result, _verbosity=Verbosity.DEFAULT):
    '''Plot statistics on the tests classified by labels if 1 or labels.

    :rtype: list(PlotTemplate)
    '''
    if len(result.test.by_labels) > 2:
        LOGGER.note('No plot for more than 2 labels.')
        return []
    if len(result.test.by_labels) == 2:
        return repr_statstestsby2labels(result)
    lok, lko, legs = ([] for i in range(3))
    for res in result.classify:
        lok.append(res['OK'])
        lko.append(res['KO'])
        legs.append(res['labels'][0])
    curve_ok = CurveElements(np.array(lok), bins=[legs], index=0, legend='OK')
    curve_ko = CurveElements(np.array(lko), bins=[legs], index=1, legend='KO')
    sbpeb = SubPlotElements(curves=[curve_ok, curve_ko],
                            axnames=[result.test.by_labels[0], '# of tests'],
                            ptype='bar')
    sbpebs = SubPlotElements(curves=[curve_ok, curve_ko],
                             axnames=[result.test.by_labels[0], '# of tests'],
                             ptype='barstack')
    return [PlotTemplate(subplots=[sbpeb, sbpebs], small_subplots=False)]



# Copyright French Alternative Energies and Atomic Energy Commission
# Contributors: valjean developers
# valjean-support@cea.fr
#
# This software is a computer program whose purpose is to analyze and
# post-process numerical simulation results.
#
# This software is governed by the CeCILL license under French law and abiding
# by the rules of distribution of free software. You can use, modify and/ or
# redistribute the software under the terms of the CeCILL license as circulated
# by CEA, CNRS and INRIA at the following URL: http://www.cecill.info.
#
# As a counterpart to the access to the source code and rights to copy, modify
# and redistribute granted by the license, users are provided only with a
# limited warranty and the software's author, the holder of the economic
# rights, and the successive licensors have only limited liability.
#
# In this respect, the user's attention is drawn to the risks associated with
# loading, using, modifying and/or developing or reproducing the software by
# the user in light of its specific status of free software, that may mean that
# it is complicated to manipulate, and that also therefore means that it is
# reserved for developers and experienced professionals having in-depth
# computer knowledge. Users are therefore encouraged to load and test the
# software's suitability as regards their requirements in conditions enabling
# the security of their systems and/or data to be ensured and, more generally,
# to use and operate it in the same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.
#
# -*- coding: utf-8 -*-
'''This module collects a few task classes that can be used with the
:mod:`~.scheduler` and :mod:`~.depgraph` modules.

This module defines a dummy :class:`Task` class that may be used as a base
class and extended.

The :meth:`.Task.do()` method takes two arguments:

* `env` is an environment for task execution. The idea of the environment is
  that tasks may use it to store information about their execution. For
  instance, a task may create a file and store its location in the environment,
  so that later tasks may be able to retrieve it.  The type of `env` is really
  immaterial, but it is probably natural to use a key-value mapping of some
  kind.  Note, however, that most of the tasks defined in the :mod:`~.task`
  module hierarchy expect `env` to be an :class:`~.Env` object.
* `config` is a :class:`~.config.Config` object describing the configuration
  for the current run. Tasks may look up global configuration values here.

The :class:`Task` class models two types of inter-task dependencies. **Hard
dependencies** represent dependencies that are crucial for the execution of the
task at hand. If task `A` has a hard dependency on task `B`, it means that `A`
cannot run unless `B` has successfully completed. If `B` fails, then it makes
no sense to run `A`.  On the other hand, if task `A` has a **soft dependency**
on task `B`, it means that `A` will not start before `B`'s termination, but it
makes sense to run `A` even if `B` fails.
'''

import enum
import json
from abc import ABC, abstractmethod
import logging


LOGGER = logging.getLogger(__name__)


#: Enumeration for the task status. The possible values are:
#:
#: * ``WAITING`` (the task is waiting to be scheduled)
#: * ``PENDING`` (the task is under execution)
#: * ``DONE`` (the task was executed and it succeeded)
#: * ``FAILED`` (the task was executed and it failed)
#: * ``SKIPPED`` (the task was skipped by the scheduler; this may happen, for
#:   instance, if the one of the task dependencies was not successful)
TaskStatus = enum.IntEnum('TaskStatus',  # pylint: disable=invalid-name
                          'WAITING PENDING DONE FAILED SKIPPED')


class TaskError(Exception):
    '''An error that may be raised by :class:`~Task` classes.'''


class Task(ABC):
    '''Base class for other task classes.'''

    def __init__(self, name, *, deps=None, soft_deps=None):
        '''Initialize the task.

        :param str name: The name of the task. Task names **must** be unique!
        :param deps: The list of (hard) dependencies for this task. It must be
                     either `None` (i.e. no dependencies) or list of
                     :class:`Task` objects.
        :type deps: list(Task) or None
        :param soft_deps: The list of soft dependencies for this task. It must
                          be either `None` (i.e. no dependencies) or list of
                          :class:`Task` objects.
        :type soft_deps: list(Task) or None
        '''
        LOGGER.debug('creating task %s', name)
        self.name = name

        self.depends_on = set()
        if deps is not None:
            if not isinstance(deps, (tuple, list, set)):
                errmsg = ('The `deps` task argument must '
                          'be either a collection of tasks or None; '
                          f'type {type(deps)} found')
                raise TypeError(errmsg)
            self.depends_on.update(deps)

        self.soft_depends_on = set()
        if soft_deps is not None:
            if not isinstance(soft_deps, (tuple, list, set)):
                errmsg = ('The `soft_deps` task argument must '
                          'be either a collection of tasks or None; '
                          f'type {type(soft_deps)} found')
                raise TypeError(errmsg)
            self.soft_depends_on.update(soft_deps)

    @abstractmethod
    def do(self, env, config):
        '''Perform a task.

        :param env: The environment for this task.
        '''
        raise NotImplementedError('do() not implemented for Task')

    def __str__(self):
        return repr(self.name)  # use repr() to add quotes around the task name

    def __repr__(self):
        return f"Task('{self.name}')"

    def add_dependency(self, dep):
        '''Add an item to the list of dependencies of this task.'''
        self.depends_on.add(dep)

    def depends(self, other):
        '''Return `True` if `self` depends on `other`.'''
        return other in self.depends_on

    def soft_depends(self, other):
        '''Return `True` if `self` has a soft dependency on `other`.'''
        return other in self.soft_depends_on


class DelayTask(Task):
    '''Task that waits for the specified number of seconds. This task is useful
    to test scheduling algorithms under different load conditions.
    '''

    def __init__(self, name, delay=1.):
        '''Initialize the task from a given delay.

        :param float delay: The amount of time (in seconds) that this task will
                            wait when executed.
        '''
        super().__init__(name)
        self.delay = float(delay)

    def do(self, env, config):
        '''Perform the task (i.e. sleep; I wish my life was like that).

        :param env: The environment. Ignored.
        '''

        from time import sleep
        LOGGER.info('DelayTask %s sleeping %f seconds...',
                    self, self.delay)
        sleep(self.delay)
        LOGGER.info('DelayTask %s waking up!', self)
        return {}, TaskStatus.DONE


def det_hash(*args):
    '''Produce a deterministic hash for the collection of objects passed as
    an argument.'''
    from hashlib import sha256
    hasher = sha256()
    for thing in args:
        LOGGER.debug('hashing: %r', thing)
        json_thing = json.dumps(thing, sort_keys=True)
        LOGGER.debug('in json: %s', json_thing)
        hasher.update(json_thing.encode('utf-8'))
    digest = hasher.hexdigest()
    LOGGER.debug('resulting hash: %s', digest)
    return digest


def close_dependency_graph(tasks):
    '''Return the tasks along with all their dependencies.

    :param list tasks: A list of tasks.
    :returns: The list of tasks, their dependencies, the dependencies of their
              dependencies and so on.
    :rtype: list(Task)
    '''
    queue = set(tasks)
    all_tasks = queue.copy()
    while queue:
        deps = set(dep for task in queue for dep in task.depends_on
                   if task.depends_on is not None)
        soft_deps = set(dep for task in queue for dep in task.soft_depends_on
                        if task.soft_depends_on is not None)
        all_tasks.update(deps)
        all_tasks.update(soft_deps)
        queue = deps | soft_deps
    return list(all_tasks)



# Copyright French Alternative Energies and Atomic Energy Commission
# Contributors: valjean developers
# valjean-support@cea.fr
#
# This software is a computer program whose purpose is to analyze and
# post-process numerical simulation results.
#
# This software is governed by the CeCILL license under French law and abiding
# by the rules of distribution of free software. You can use, modify and/ or
# redistribute the software under the terms of the CeCILL license as circulated
# by CEA, CNRS and INRIA at the following URL: http://www.cecill.info.
#
# As a counterpart to the access to the source code and rights to copy, modify
# and redistribute granted by the license, users are provided only with a
# limited warranty and the software's author, the holder of the economic
# rights, and the successive licensors have only limited liability.
#
# In this respect, the user's attention is drawn to the risks associated with
# loading, using, modifying and/or developing or reproducing the software by
# the user in light of its specific status of free software, that may mean that
# it is complicated to manipulate, and that also therefore means that it is
# reserved for developers and experienced professionals having in-depth
# computer knowledge. Users are therefore encouraged to load and test the
# software's suitability as regards their requirements in conditions enabling
# the security of their systems and/or data to be ensured and, more generally,
# to use and operate it in the same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

'''Module containing all available methods to convert a test result in a table
to be converted in rst.
'''
import logging
from itertools import chain
import numpy as np
from ..cosette.task import TaskStatus
from ..gavroche.diagnostics.stats import TestOutcome, classification_counts
from .templates import TableTemplate, TextTemplate
from .verbosity import Verbosity

# turn off pylint warnings about invalid names in this file; there are just too
# many long function names and they cannot be renamed because
# javert.representation looks for them by programmatically constructing their
# name based on the name of the test result class, the verbosity, etc.
# pylint: disable=invalid-name


LOGGER = logging.getLogger(__name__)


def repr_bins(dsref):
    '''Representation of bins in tables.

    When bins are given by edges, representation is ``min - max``, when they
    are given at center, representation is ``center``.

    Trivial dimensions are not represented, i.e. dimensions where there is only
    one bin.

    If there are more than one non-trivial dimensions, some repetition is
    expected. For example with two non-trivial dimensions of two bins each one
    point will be defined by its coordinated in the two dimensions and we
    expect all the bins to possibily be shown in a table. We expected 4 bins
    and their associated values in that case.

    Let's consider the following dataset:
    >>> from valjean.eponine.dataset import Dataset
    >>> import numpy as np
    >>> from collections import OrderedDict

    >>> vals = np.arange(6).reshape(1, 2, 1, 3, 1)
    >>> errs = np.array([0.1]*6).reshape(1, 2, 1, 3, 1)
    >>> bins = OrderedDict([('bacon', np.array([0, 1])),
    ...                     ('egg', np.array([0, 2, 4])),
    ...                     ('sausage', np.array([10, 20])),
    ...                     ('spam', np.array([-5, 0, 5])),
    ...                     ('tomato', np.array([-2, 2]))])
    >>> ds = Dataset(vals, errs, bins=bins)
    >>> names, rbins = repr_bins(ds)
    >>> print(list(ds.bins.keys()))
    ['bacon', 'egg', 'sausage', 'spam', 'tomato']
    >>> print(names)
    ['egg', 'spam']
    >>> print(ds.shape)
    (1, 2, 1, 3, 1)
    >>> print([rb.shape for rb in rbins])
    [(1, 2, 1, 3, 1), (1, 2, 1, 3, 1)]

    ``'bacon'`` and ``'sausage'`` are trivial dimensions, so won't be
    represented in the table, but we expect 6 values corresponding to 2 bins in
    ``'egg'`` and 3 in ``'spam'``. Each value corresponds to a line in the
    table so each columns should have the same size and the same shape, the
    shape of the given dataset without trivial dimensions. We then have 3
    ``'spam'`` bins in each ``'egg'`` bins or 2 ``'egg'`` bins in each
    ``'spam'`` bins. Each couple appears only once.

    >>> for name, rbin in zip(names, rbins):
    ...     print(name, ':', rbin.flatten())
    egg : ['0 - 2' '0 - 2' '0 - 2' '2 - 4' '2 - 4' '2 - 4']
    spam : ['-5' '0' '5' '-5' '0' '5']

    As expected from the bins, ``'egg'`` bins are given by edges
    (``min - max``) while ``'spam'`` bins are given by center (``center``).

    :param dsref: dataset
    :type dsref: Dataset
    :returns: list of the non-trivial dimensions and a tuple of the bins
    :rtype: (list(str), tuple(numpy.ndarray))

    The tuple must have the same length as the list of dimensions and the bins
    inside must have the same shape as ``dsref.value``.
    '''
    bins, dim_names = [], []
    for idim, dim in enumerate(dsref.bins.keys()):
        # reject trivial dimensions (1 bin or no bin)
        if dsref.value.shape[idim] < 2:
            continue
        dim_names.append(dim)
        shape = [dsref.value.shape[idim]] + [1] * (dsref.ndim - 1 - idim)
        if dsref.value.shape[:idim] == tuple([1]*idim):
            shape = [1]*idim + shape
        if dsref.bins[dim].size == dsref.value.shape[idim]+1:
            dbins = [f"{a:.4g} - {b:.4g}"
                     for a, b in zip(dsref.bins[dim][:-1],
                                     dsref.bins[dim][1:])]
        else:
            dbins = ([f"{a:.4g}" for a in dsref.bins[dim]]
                     if dsref.bins[dim].dtype.kind != 'U'
                     else dsref.bins[dim])
        bins.append(np.array(dbins).reshape(shape))
    bbins = np.broadcast_arrays(*bins)
    return dim_names, tuple(bbins)


def repr_testresultequal(result, verbosity=Verbosity.DEFAULT):
    '''Represent the result of a :class:`~.TestEqual` test.

    :param TestResultEqual result: a test result.
    :param Verbosity verbosity: verbosity level
    :returns: list of templates representing a :class:`~.TestResultEqual`
    '''
    LOGGER.debug("IN repr_testresultequal, %s, res = %s",
                 verbosity, bool(result))
    if bool(result):
        if verbosity != Verbosity.FULL_DETAILS:
            return []
        return repr_equal(result)
    if verbosity.value < Verbosity.DEFAULT.value:
        return repr_equal_summary(result)
    return repr_equal(result)


def repr_equal_summary(result):
    '''Function to generate a summary table for the equal test (only tells if
    the test was successful or not).

    :param TestResultEqual result: a test result.
    :rtype: list(TextTemplate)
    '''
    LOGGER.debug("repr_equal_summary found")
    if result:
        return [TextTemplate('Equal test: OK\n\n')]
    return [TextTemplate('.. role:: hl\n\n'
                         'Equal test: :hl:`KO`\n\n')]


def repr_equal(result):
    '''Representation of equal test.

    :param TestResultEqual result: a test result.
    :returns: Representation of a :class:`~.TestResultEqual` as a table.
    :rtype: list(TableTemplate)
    '''
    LOGGER.debug("In repr_equal")
    nbins, bins = repr_bins(result.test.dsref)
    dscols = tuple((ds.value, eq)
                   for ds, eq in zip(result.test.datasets, result.equal))
    heads = nbins + [result.test.dsref.name]
    for ds in result.test.datasets:
        heads.extend([ds.name,
                      f'equal({ds.name})?'
                      if len(result.test.datasets) > 1 else 'equal?'])
    falses = np.full_like(result.test.dsref.value, False, dtype=bool)
    highlights = [falses] * (len(nbins) + 1)
    for equal in result.equal:
        highlights.extend([falses, np.logical_not(equal)])
    table_template = TableTemplate(
        *bins, result.test.dsref.value, *chain.from_iterable(dscols),
        highlights=highlights, headers=heads)
    return [table_template]


def repr_testresultapproxequal(result, verbosity=Verbosity.DEFAULT):
    '''Represent the result of a :class:`~.TestApproxEqual` test.

    :param TestResultApproxEqual result: a test result.
    :param Verbosity verbosity: verbosity level
    :returns: list of templates representing a :class:`~.TestResultApproxEqual`
    '''
    if verbosity == Verbosity.SILENT and bool(result):
        return []
    if verbosity == Verbosity.SUMMARY:
        return repr_approx_equal_summary(result)
    return repr_approx_equal(result)


def repr_approx_equal_summary(result):
    '''Function to generate a summary table for the approx equal test (only
    tells if the test was successful or not).

    :param TestResultApproxEqual result: a test result.
    :rtype: list(TextTemplate)
    '''
    LOGGER.debug("repr_approx_equal_summary found")
    if result:
        return [TextTemplate('Approx equal test: OK\n\n')]
    return [TextTemplate('.. role:: hl\n\n'
                         'Approx equal test: :hl:`KO`\n\n')]


def repr_approx_equal(result):
    '''Representation of approx equal test.

    :param TestResultApproxEqual result: a test result.
    :returns: Representation of a :class:`~.TestResultApproxEqual` as a table.
    :rtype: list(TableTemplate)
    '''
    LOGGER.debug("In repr_approx_equal")
    nbins, bins = repr_bins(result.test.dsref)
    dscols = tuple((ds.value, eq)
                   for ds, eq in zip(result.test.datasets,
                                     result.approx_equal))
    heads = nbins + [result.test.dsref.name]
    for ds in result.test.datasets:
        heads.extend([ds.name,
                      f'approx equal({ds.name})?'
                      if len(result.test.datasets) > 1
                      else 'approx equal?'])
    falses = np.full_like(result.test.dsref.value, False, dtype=bool)
    highlights = [falses] * (len(nbins) + 1)
    for approx_equal in result.approx_equal:
        highlights.extend([falses, np.logical_not(approx_equal)])
    table_template = TableTemplate(
        *bins, result.test.dsref.value, *chain.from_iterable(dscols),
        highlights=highlights, headers=heads)
    return [table_template]


def repr_testresultstudent(result, verbosity=Verbosity.DEFAULT):
    '''Represent the result of a :class:`~.TestStudent` test.

    :param TestResultStudent result: a test result.
    :param Verbosity verbosity: verbosity level
    :returns: list of templates representing a :class:`~.TestResultStudent`
    '''
    LOGGER.debug("student repr, %s, res = %s", verbosity, bool(result))
    if verbosity == Verbosity.SILENT:
        return []
    if verbosity == Verbosity.SUMMARY:
        return repr_student_summary(result)
    if verbosity in (Verbosity.DEFAULT, Verbosity.INTERMEDIATE):
        return repr_student_intermediate(result)
    return repr_student(result)


def _student_heads(test_res, bin_names):
    '''Build the column names for Student test representation.

    Reference appears first then all the tested datasets. They are identified
    by their name if unique, else their index in the list of datasets is used.

    :param TestStudent test_res: the test (not the result)
    :param list(str) bin_names: names of the bins (dimensions)
    :returns: list of headers
    '''
    heads = ['v('+test_res.dsref.name+')', 'σ('+test_res.dsref.name+')']
    for _ds in test_res.datasets:
        heads.extend(['v('+_ds.name+')', 'σ('+_ds.name+')'])
        if len(test_res.datasets) > 1:
            heads.extend([r't('+_ds.name+')', 'Student('+_ds.name+')?'])
        else:
            heads.extend(['t', 'Student?'])
    heads = bin_names + heads
    return heads


def repr_student(result):
    '''Representation of Student test result.

    :param TestResultStudent result: a test result.
    :returns: Representation of a :class:`~.TestResultStudent` as a table.
    :rtype: list(TableTemplate)
    '''
    LOGGER.debug("In repr_student")
    oracles = result.oracles()
    nbins, bins = repr_bins(result.test.dsref)
    dscols = tuple((ds.value, ds.error, tstud, studbool)
                   for ds, tstud, studbool in zip(result.test.datasets,
                                                  result.tstud,
                                                  oracles))
    falses = np.full_like(result.test.dsref.value, False, dtype=bool)
    heads = _student_heads(result.test, nbins)
    highlights = [falses]*(len(nbins) + 2)
    for oracle in oracles:
        highlights += [falses, falses, falses, np.logical_not(oracle)]
    table_template = TableTemplate(
        *bins, result.test.dsref.value, result.test.dsref.error,
        *chain.from_iterable(dscols),
        highlights=highlights, headers=heads)
    return [table_template]


def repr_student_silent(_result):
    '''Function to generate a silent table for the Student test (only tells if
    the test was successful or not).

    :param TestResultStudent result: a Student test result.
    :returns: empty list
    '''
    LOGGER.debug("student silent")
    return []


def repr_student_summary(result):
    '''Function to generate a summary table for the Student test (only tells if
    the test was successful or not).

    :param TestResultStudent result: a Student test result.
    :rtype: list(TextTemplate)
    '''
    LOGGER.debug("repr_student_summary found")
    if result:
        return [TextTemplate('Student test: OK\n\n')]
    return [TextTemplate('.. role:: hl\n\n'
                         'Student test: :hl:`KO`\n\n')]


def repr_student_intermediate(result):
    '''Function to generate an intermediate table for the Student test: print
    all the failing results.

    :param TestResultStudent result: a Student test result.
    :rtype: list(TableTemplate)
    '''
    LOGGER.debug("repr_student_intermediate found")
    if result:
        return repr_student_summary(result)
    if result.test.dsref.shape == ():
        return repr_student(result)
    oracles = result.oracles()
    falses_ind = np.ones_like(result.test.dsref.value)
    for oracle in oracles:
        falses_ind[np.where(oracle == 0)] = 0
    nbins, bins = repr_bins(result.test.dsref)
    dscols = tuple((ds.value[np.where(falses_ind == 0)],
                    ds.error[np.where(falses_ind == 0)],
                    tstud[np.where(falses_ind == 0)],
                    studbool[np.where(falses_ind == 0)])
                   for ds, tstud, studbool in zip(result.test.datasets,
                                                  result.tstud,
                                                  oracles))
    heads = _student_heads(result.test, nbins)
    falses = np.full_like(result.test.dsref.value[np.where(falses_ind == 0)],
                          False, dtype=bool)
    highlights = [falses]*(len(nbins) + 2)
    for oracle in oracles:
        highlights += [falses, falses, falses,
                       np.logical_not(oracle[np.where(falses_ind == 0)])]
    cols = (list((b[np.where(falses_ind == 0)],) for b in bins)
            + [(result.test.dsref.value[np.where(falses_ind == 0)],
                result.test.dsref.error[np.where(falses_ind == 0)])])
    table_template = TableTemplate(
        *chain.from_iterable(cols), *chain.from_iterable(dscols),
        highlights=highlights, headers=heads)
    return [table_template]


def repr_testresultbonferroni(result, verbosity=Verbosity.DEFAULT):
    '''Represent the result of a :class:`~.TestBonferroni` test.

    Only represents the Bonferroni result, not the input test result.

    :param TestResultBonferroni result: a test result.
    :param Verbosity verbosity: verbosity level
    :returns: list of templates representing a :class:`~.TestResultBonferroni`
    '''
    LOGGER.debug("bonf repr, %s, res = %s", verbosity, bool(result))
    if verbosity == Verbosity.SILENT and bool(result):
        return []
    if verbosity == Verbosity.SUMMARY:
        return repr_bonferroni_summary(result)
    return repr_bonferroni(result)


def repr_bonferroni(result):
    '''Reprensetation of Bonferroni test result.

    Only represents the Bonferroni result, not the input test result.

    :param TestResultBonferroni result: a test result.
    :returns: Representation of a :class:`~.TestResultBonferroni` as a table.
    :rtype: list(TableTemplate)
    '''
    ndatasets = len(result.first_test_res.test.datasets)
    oracles = list(result.oracles())
    highlights = [[False] * ndatasets] * 5  # 5 non-highlighted columns
    highlights.append([not oracle for oracle in oracles])
    table_template = TableTemplate(
        [f"{result.first_test_res.test.dsref.name} vs {dtest.name}"
         for dtest in result.first_test_res.test.datasets],
        [result.test.ntests] * ndatasets,
        [result.test.alpha] * ndatasets,
        [result.test.bonf_signi_level] * ndatasets,
        [min(pval) for pval in result.first_test_res.pvalue],
        oracles,
        highlights=highlights,
        headers=['test', 'ndf', 'α', 'α(Bonferroni)', 'min(p-value)',
                 'Bonferroni?'])
    return [table_template]


def repr_bonferroni_summary(result):
    '''Represent the result of a :class:`~.TestBonferroni` test for the
    SUMMARY level of verbosity.

    :param TestResultBonferroni result: a test result.
    :returns: Representation of a :class:`~.TestResultBonferroni` as a table.
    :rtype: list(TextTemplate)
    '''
    if result:
        return [TextTemplate('Bonferroni test: OK\n\n')]
    return [TextTemplate('.. role:: hl\n\n'
                         'Bonferroni test: :hl:`KO`\n\n')]


def repr_testresultholmbonferroni(result, verbosity=Verbosity.DEFAULT):
    '''Represent the result of a :class:`~.TestHolmBonferroni` test.

    :param TestResultHolmBonferroni result: a test result.
    :param Verbosity verbosity: verbosity level
    :returns: list of templates representing a
        :class:`~.TestResultHolmBonferroni`
    '''
    LOGGER.debug("HB res, %s, res = %s", verbosity, bool(result))
    if verbosity == Verbosity.SILENT:
        if bool(result):
            return []
        return repr_holm_bonferroni_summary(result)
    if verbosity == Verbosity.SUMMARY:
        return repr_holm_bonferroni_summary(result)
    return repr_holm_bonferroni(result)


def repr_holm_bonferroni(result):
    '''Reprensetation of Holm-Bonferroni test result.

    Only represents the Holm-Bonferroni result, not the input test result.

    :param TestResultHolmBonferroni result: a test result.
    :returns: Representation of a :class:`~.TestResultHolmBonferroni` as a
        table.
    :rtype: list(TableTemplate)
    '''
    ndatasets = len(result.first_test_res.test.datasets)
    oracles = list(result.oracles())
    highlights = [[False] * ndatasets] * 6  # 6 non-highlighted columns
    highlights.append([not oracle for oracle in oracles])
    table_template = TableTemplate(
        [f"{result.first_test_res.test.dsref.name} vs {dtest.name}"
         for dtest in result.first_test_res.test.datasets],
        [result.test.ntests] * ndatasets,
        [result.test.alpha] * ndatasets,
        [np.amin(pval) for pval in result.first_test_res.pvalue],
        [np.amin(alpha_i) for alpha_i in result.alphas_i],
        list(result.nb_rejected),
        oracles,
        highlights=highlights,
        headers=['test', 'ndf', 'α', 'min(p-value)', 'min(α)',
                 'N rejected', 'Holm-Bonferroni?'])
    return [table_template]


def repr_holm_bonferroni_summary(result):
    '''Represent the result of a :class:`~.TestHolmBonferroni` test for the
    SUMMARY level of verbosity.

    :param TestResultHolmBonferroni result: a test result.
    :returns: Representation of a :class:`~.TestResultHolmBonferroni` as a
        table.
    :rtype: list(TextTemplate)
    '''
    if result:
        return [TextTemplate('Holm-Bonferroni test: OK\n\n')]
    return [TextTemplate('.. role:: hl\n\n'
                         'Holm-Bonferroni test: :hl:`KO`\n\n')]


def percent_fmt(num, den):
    '''Format a fraction as a percentage. Example:

    >>> percent_fmt(2, 4)
    '2/4 (50.0%)'
    >>> percent_fmt(0, 3)
    '0/3 (0.0%)'
    >>> percent_fmt(7, 7)
    '7/7 (100.0%)'
    >>> percent_fmt(0, 0)
    '0/0 (???%)'

    :param int num: the numerator.
    :param int den: the denominator.
    :rtype: str
    '''
    nbsp = ' '
    if den != 0:
        percent = 100.0 * num / den
        return f'{num:d}/{den:d}{nbsp}({percent:.1f}%)'
    return f'{num:d}/{den:d}{nbsp}(???%)'


def repr_testresultstatstasks(result, verbosity=Verbosity.DEFAULT):
    '''Represent a :class:`~.TestResultStatsTasks` as a table. The table
    breaks down the tasks by status.

    :param TestResultStatsTasks result: the test result to represent.
    :param Verbosity verbosity: verbosity level
    :returns: list of templates representing the test result.
    '''
    LOGGER.debug("In repr_testresultstatstasks")
    if verbosity == Verbosity.SILENT and bool(result):
        return []
    return repr_testresultstats(result, TaskStatus.DONE, 'tasks')


def repr_testresultstatstests(result, verbosity=Verbosity.DEFAULT):
    '''Represent a :class:`~.TestResultStatsTests` as a table. The table
    breaks down the tests by success status.

    :param TestResultStatsTests result: the test result to represent.
    :param Verbosity verbosity: verbosity level
    :returns: the tables representing the test result.
    :rtype: list(TableTemplate)
    '''
    LOGGER.debug("In repr_testresultstatstests")
    if verbosity == Verbosity.SILENT and bool(result):
        return []
    return repr_testresultstats(result, TestOutcome.SUCCESS, 'tests')


def repr_testresultstats(result, status_ok, label):
    '''Helper function for :func:`repr_testresultstatstests` and
    :func:`repr_testresultstatstasks`. It generates a table with the
    `status_ok` value in the first row. Non-null results in other rows are
    considered as failures, and are highlighted if the count is non-zero. Null
    results are omitted from the table.

    :param result: the test result to represent.
    :type result: TestResultStatsTasks or TestResultStatsTests
    :param status_ok: the status value that must be considered as a success.
    :param str label: the type of things that we are testing (``'tests'`` or
        ``'tasks'``)
    :returns: the tables representing the test result.
    :rtype: list(TableTemplate)
    '''
    classify = result.classify
    statuses, counts = classification_counts(classify, status_ok)
    statuses_txt = [status.name for status in statuses]
    n_tasks = sum(counts)
    statuses_txt.append('total')
    counts.append(n_tasks)
    percents = [percent_fmt(count, n_tasks) for count in counts]

    hl_column = [False]
    hl_column.extend(count > 0 for count in counts[1:-1])
    table = TableTemplate(statuses_txt, percents, headers=['status', 'counts'],
                          highlights=[hl_column, hl_column])
    text = []
    for status, status_txt in zip(statuses, statuses_txt):
        if status_ok == status:
            continue
        text.append(f'List of {label} with status {status_txt}:\n\n')
        for item in sorted(classify[status]):
            if item.fingerprint:
                text.append(f'#. :ref:`{item.name} '
                            f'<anchor_{item.fingerprint}>`')
            else:
                text.append(f'#. {item.name}')
        text.append('\n')
    return [table, TextTemplate('\n'.join(text))]


def repr_testresultstatstestsbylabels(result, verbosity=Verbosity.DEFAULT):
    '''Represent a :class:`~.TestResultStatsTestsByLabels` as tables. Shape of
    the table may change according to the number of flags required.

    :param TestResultStatsTestsByLabels result: the test result to represent.
    :param Verbosity verbosity: verbosity level
    :returns: the tables representing the test result.
    :rtype: list(TableTemplate)
    '''
    LOGGER.debug("In repr_testresultstatstestsbylabels")
    if verbosity == Verbosity.SILENT and bool(result):
        return []
    if verbosity == Verbosity.SUMMARY:
        return repr_testresultstatsbylabels_summary(result)
    return repr_testresultstatsbylabels(result)


def _sbl_labels_1column(labels, results, oracles):
    '''Build a :class:`TableTemplate` from the results where labels are
    concantenated in one column.

    :param tuple(str) labels: labels used to make statistics
    :param dict results: result from :class:`TestResultStatsTestsByLabels`
    :param list(bool) oracles: result of each individual test
    :rtype: TableTemplate
    '''
    lnames, lok, lko, hlight = [], [], [], []
    for res, ora in zip(results, oracles):
        lnames.append('/'.join(res['labels']))
        lok.append(percent_fmt(res['OK'], res['total']))
        lko.append(percent_fmt(res['KO'], res['total']))
        hlight.append([not ora])
    return TableTemplate(
        lnames, lok, lko,
        headers=['/'.join(labels), r'% success', r'% failure'],
        highlights=[hlight]*3)


def _sbl_1colbylabel(labels, results, oracles):
    '''Build a :class:`TableTemplate` from the results where each label will be
    represented in a separate column.

    :param tuple(str) labels: labels used to make statistics
    :param dict results: result from :class:`TestResultStatsTestsByLabels`
    :param list(bool) oracles: result of each individual test
    :rtype: TableTemplate
    '''
    lnames = [[] for _ in range(len(labels))]
    lok, lko, hlight = [], [], []
    for res, ora in zip(results, oracles):
        for i, lab in enumerate(res['labels']):
            lnames[i].append(lab)
        lok.append(percent_fmt(res['OK'], res['total']))
        lko.append(percent_fmt(res['KO'], res['total']))
        hlight.append([not ora])
    return TableTemplate(
        *lnames, lok, lko,
        headers=[*labels] + [r'% success', r'% failure'],
        highlights=[hlight]*(len(labels)+2))


def repr_testresultstatsbylabels(result):
    '''Function to print detailed statistics on tests, per category and sample
    run.

    :param TestResultStatsTestsByLabels result: the test result to represent.
    :returns: the tables representing the test result.
    :rtype: list(TableTemplate)
    '''
    LOGGER.debug("In repr_testresultstatsbylabels")
    res = []
    res.append(_sbl_1colbylabel(
        result.test.by_labels, result.classify, result.oracles()))
    if result.nb_missing_labels() != 0:
        res.append(TextTemplate('At least one of the labels used for sorting '
                                f'{result.test.by_labels} is missing in '
                                f'{result.nb_missing_labels()} tests\n\n'))
    return res


def repr_testresultstatsbylabels_summary(result):
    '''Function to print detailed statistics on tests, per category and sample
    run in summary case: only print failed cases.

    :param TestResultStatsTestsByLabels result: the test result to represent.
    :returns: the tables representing the test result.
    :rtype: list(TableTemplate), list(TextTemplate)
    '''
    LOGGER.debug("In repr_testresultstatsbylabels")
    res = []
    lnames = [[] for _ in range(len(result.test.by_labels))]
    lok, lko, hlight = [], [], []
    for resu, ora in zip(result.classify, result.oracles()):
        if ora:
            continue
        for i, lab in enumerate(resu['labels']):
            lnames[i].append(lab)
        lok.append(percent_fmt(resu['OK'], resu['total']))
        lko.append(percent_fmt(resu['KO'], resu['total']))
        hlight.append([not ora])
    tabtemp = TableTemplate(
        *lnames, lok, lko,
        headers=[*result.test.by_labels] + [r'% success', r'% failure'],
        highlights=[hlight]*(len(result.test.by_labels)+2))
    if lok:
        res.append(tabtemp)
    else:
        res.append(TextTemplate('No failed test found.\n\n'))
    if result.nb_missing_labels() != 0:
        res.append(TextTemplate('At least one of the labels used for sorting '
                                f'{result.test.by_labels} is missing in '
                                f'{result.nb_missing_labels()} tests\n\n'))
    return res


def repr_testresultmetadata(result, verbosity=Verbosity.DEFAULT):
    '''Represent the result of a :class:`~.TestMetadata` test.

    :param TestResultMetadata result: a test result.
    :param Verbosity verbosity: verbosity level
    :returns: list of templates representing a :class:`~.TestResultMetadata`
    '''
    LOGGER.debug("metadata res, %s, res = %s", verbosity, bool(result))
    if verbosity == Verbosity.SILENT:
        return []
    if verbosity == Verbosity.SUMMARY:
        return repr_metadata_summary(result)
    if verbosity == Verbosity.DEFAULT:
        return repr_metadata(result)
    if verbosity == Verbosity.INTERMEDIATE:
        return repr_metadata_intermediate(result)
    if verbosity.value >= Verbosity.FULL_DETAILS.value:
        return repr_metadata_full_details(result)
    return repr_metadata(result)


def repr_metadata(result):
    '''Function to generate a table from the metadata test results.

    :param TestResultMetadata result: a test result.
    :rtype: list(TableTemplate)
    '''
    if result:
        return [TableTemplate(["Metadata:"], ["OK"],
                              highlights=[[False], [False]])]
    ko_list = list(result.only_failed_comparisons().keys())
    return [TableTemplate(["Failed metadata:"], [', '.join(ko_list)],
                          highlights=[[False], [True]])]


def repr_metadata_full_details(result):
    '''Function to generate a table from the metadata test results.

    :param TestResultMetadata result: a test result.
    :rtype: list(TableTemplate)
    '''
    LOGGER.debug("repr_metadata_full_details")
    samp_names = list(result.test.dmd)
    keys = []
    tdict = {name: [] for name in samp_names}
    hdict = {name: [] for name in samp_names}
    for dkey in result.test.all_md.keys():
        keys.append(dkey)
        for nam in samp_names:
            tdict[nam].append(str(result.test.all_md[dkey][nam]))
            hdict[nam].append(not result.dict_res[dkey][nam])
    ocols = list(tdict.values())
    highl = [[False]*len(keys)]
    highl.extend(hdict.values())
    heads = ['key']
    heads.extend(result.test.dmd)
    table = TableTemplate(keys, *ocols,
                          highlights=highl,
                          headers=heads)
    return [table]


def repr_metadata_intermediate(result):
    '''Function to generate a table from the metadata test results.

    :param TestResultMetadata result: a test result.
    :rtype: list(TableTemplate)
    '''
    LOGGER.debug("repr_metadata_intermediate")
    if not result.only_failed_comparisons():
        return repr_metadata_summary(result)
    samp_names = list(result.test.dmd)
    keys = []
    tdict = {name: [] for name in samp_names}
    hdict = {name: [] for name in samp_names}
    for dkey in result.only_failed_comparisons():
        keys.append(dkey)
        for nam in samp_names:
            tdict[nam].append(str(result.only_failed_comparisons()[dkey][nam]))
            hdict[nam].append(not result.dict_res[dkey][nam])
    ocols = list(tdict.values())
    highl = [[False]*len(keys)]
    highl.extend(hdict.values())
    heads = ['key']
    heads.extend(result.test.dmd)
    table = TableTemplate(keys, *ocols,
                          highlights=highl,
                          headers=heads)
    return [table]


def repr_metadata_summary(result):
    '''Function to generate a table from the metadata test results.

    :param TestResultMetadata result: a test result.
    :rtype: list(TextTemplate)
    '''
    LOGGER.debug("repr_metadata_summary")
    if result:
        return [TextTemplate('Metadata: OK\n\n')]
    return [TextTemplate('.. role:: hl\n\n'
                         'Metadata test: :hl:`KO`\n\n')]


def repr_metadata_silent(_result):
    '''Function to generate a table from the metadata test results.

    :param TestResultMetadata result: a test result.
    :returns: empty list
    '''
    LOGGER.debug("repr_metadata_silent")
    return []


def repr_testresultexternal(_result, _verbosity=Verbosity.DEFAULT):
    '''Represent external test as tables -> no table done.

    If tables are required they are already done. Their representation in the
    report is done by :class:`.ExternalRepresenter`.

    :returns: empty list
    '''
    return []


def repr_testresultfailed(result, _verbosity=Verbosity.DEFAULT):
    '''Represent a failed result as rst text.

    :param TestResultFailed result: a failed test result.
    :rtype: list(TextTemplate)
    '''
    LOGGER.debug("In repr_testresultfailed")
    defmsg = ' failed with message:'
    cname = result.test.__class__.__name__
    text = f':hl:`{cname}{defmsg}`\n{result.msg}\n'
    return [TextTemplate('.. role:: hl\n\n' + text.replace('\n', '\n\n'))]
